/*
 *  login.c
 *
 *  Some functions to log in server.
 *
 *  Copyright (C) 2008  Huang Guan
 *
 *  2008-01-31 Created.
 *
 *  Description: This file mainly includes the functions about 
 *               Getting Login Token
 *               Sending and Handling Login Packet
 *               Changing User Status.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#ifdef __WIN32__
#include <winsock.h>
#include <wininet.h>
#else
#include <sys/socket.h>
#include <arpa/inet.h>
#endif

#include <unistd.h>
#include "myqq.h"
#include "protocol.h"
#include "debug.h"
#include "mysocket.h"
#include "crypt.h"
#include "md5.h"
#include "cs.h"

typedef struct __attribute__((packed)) myqq_prot_login_redirect_s {
	char		ret;
	uint		uid;
	uint		server_ip;
	ushort		server_port;
}myqq_prot_login_redirect_t;

typedef struct __attribute__((packed)) myqq_prot_login_info_s {
	char		ret;
	char		session_key[16];
	uint		uid;
	uint		client_ip;	/* those detected by server */
	ushort		client_port;
	uint 		server_ip;
	ushort		server_port;
	time_t		login_time;
	char		unknown1[26];
	uint		unknown_server1_ip;
	ushort		unknown_server1_port;
	char		unknown_server2_ip[4];
	ushort		unknown_server2_port;
	ushort		unknown2;	/* 0x0001 */
	ushort		unknown3;	/* 0x0000 */
	char		unknown4[32];
	char		unknown5[12];
	uint		last_client_ip;
	time_t		last_login_time;
	char		unknown6[8];
}myqq_prot_login_info_t;

//ȡ¼ơ
static int login_get_token()
{
	int len_in, i;
	myqq_prot_packet_t *packet_reply = packet_alloc( 256, PACKET_IN );
	i = 0;
	if( cs_send_cmd( QQ_CMD_REQUEST_LOGIN_TOKEN, (char*)&i, 1, 0 ) < 0 )
		return -EFAULT;
	DBG("1. GET LOGIN TOKEN PACKET SENT.");
	//wait for reply
	for( i=0; i<8; i++ )
	{
		len_in = packet_recv( packet_reply, 256 );
		if( packet_match( packet_reply ) )
		{
			DBG("Not the packet I want.");
			continue;
		}
		break;
	}
	if( packet_reply->data_in[0] != 0 ) //failed
	{
		DBG("incorrect data");
		return -EFAULT;
	}
	//save the key token
	myqq.token_len = packet_reply->data_in[1];
	memcpy( myqq.token, &packet_reply->data_in[2], myqq.token_len );
	//free all, will free packet_out too.
	packet_free( packet_reply );
	return 0;
}

//Use Another Server
static int login_redirect( char* data, int len )
{
	struct myqq_prot_login_redirect_s* red = 
		(struct myqq_prot_login_redirect_s*) data;
	if( ntohl(red->uid) != myqq.uid )
	{
		DBG("Error: red->uid: %d", ntohl(red->uid) ); 
	}
	myqq.server_ip = ntohl( red->server_ip );
	myqq.server_port = ntohs( red->server_port );
	return 2; //demand relogin
}

//save login info
static void login_save_info( char* data, int len )
{
	if( sizeof(myqq_prot_login_info_t)!=len )
	{
		DBG("WARNING: Size not matched. %d and %d", sizeof(myqq_prot_login_info_t), len );
		///hex_dump( data, len );
	}
	myqq_prot_login_info_t * info = (myqq_prot_login_info_t*) data;
	memcpy( myqq.session_key, info->session_key, 16 );
	//session_md5
	char res[20];
	md5_state_t mst;
	memcpy( res, &myqq.uid, 4 );
	memcpy( res + 4, myqq.session_key, 16 );
	md5_init( &mst );
	md5_append( &mst, (md5_byte_t*)res, 20 );
	md5_finish( &mst, (md5_byte_t*)myqq.session_md5 ); //an important key for im and file transfering
	
	if( myqq.uid != ntohl(info->uid) ) 
		DBG("uid not equal.");
	myqq.server_ip = ntohl( info->server_ip );
	myqq.server_port = ntohs( info->server_port );
	myqq.client_ip = ntohl( info->client_ip );
	myqq.client_port = ntohs( info->client_port );
	myqq.last_client_ip = ntohl( info->last_client_ip );
	myqq.login_time = ntohl( info->login_time );
	myqq.last_login_time = ntohl( info->last_login_time );
	DBG("Last login time: %s", ctime( &myqq.last_login_time ) );
}

//͵¼Ϣ
static int login_send_info()
{
	int len;
	char data[1024];	//final data
	memset( data, 0, 1024 );
	//ʼ¼Կ
	memset( myqq.session_key,0x01, 16 );
	memcpy( data, myqq.session_key, 16 );
	//
	md5_state_t mst;
	char res[16], tmp[800], data_encrypted[800];
	memset( tmp, 0, 800 ); //initialize the temporary data
	md5_init( &mst );
	md5_append( &mst, (md5_byte_t*)myqq.password, strlen(myqq.password) );
	md5_finish( &mst, (md5_byte_t*)res );
	//ڶ
	md5_init( &mst );
	md5_append( &mst, (md5_byte_t*)res, 16 );
	md5_finish( &mst, (md5_byte_t*)myqq.pass_encrypted );
	//һַ,Ķԡtmp
	//tmp: 0-15
	qq_crypt( ENCRYPT, (unsigned char*)"", 0, 
		(unsigned char*)myqq.pass_encrypted, (unsigned char*)tmp, &len); //len==16
	char* curp = tmp + 16;
	//tmp: 16-51
	static const char raw1[36] = { 
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 			0x00,0x00,0x00,0x86,0xcc,0x4c,0x35,0x2c,	0xd3,0x73,0x6c,0x14,0xf6,0xf6,0xaf,0xc3,
		0xfa,0x33,0xa4,0x01 }; //36bytes
	memcpy( curp, raw1, 36 );
	curp += 36;
	//tmp: 52
	if( (myqq.mode & 0xFF) != LM_HIDDEN ) //not hidden mode
	{
		*curp = 0x0a;
	}else{
		*curp = 0x28;
		DBG("Want to hide??");
	}
	curp ++;
	//tmp: 53-68
	static const char raw2[16] = {
		0x8d, 0x8b, 0xfa,0xec, 0xd5, 0x52, 0x17, 0x4a, 0x86, 0xf9, 0xa7,
		0x75, 0xe6, 0x32, 0xd1, 0x6d
		};
	memcpy( curp, raw2, 16 );
	curp += 16;
	//tmp: 69
	*curp = myqq.token_len;
	curp ++;
	//tmp: 70-93
	memcpy( curp, myqq.token, myqq.token_len );
	curp += myqq.token_len;
	//tmp: rest
	static const char raw3[] = {
		0x40,0x0b,0x04,0x02,0x00,0x01,0x00,0x00,	0x00,0x00,0x00,0x03,0x09,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,0x01,0xe9,0x03,0x01,	0x00,0x00,0x00,0x00,0x00,0x01,0xf3,0x03,
		0x00,0x00,0x00,0x00,0x00,0x00,0x01,0xed,	0x03,0x00,0x00,0x00,0x00,0x00,0x00,0x01,
		0xec,0x03,0x00,0x00,0x00,0x00,0x00,0x00,	0x03,0x05,0x00,0x00,0x00,0x00,0x00,0x00,
		0x00,0x03,0x07,0x00,0x00,0x00,0x00,0x00,	0x00,0x00,0x01,0xee,0x03,0x00,0x00,0x00,
		0x00,0x00,0x00,0x01,0xef,0x03,0x00,0x00,	0x00,0x00,0x00,0x00,0x01,0xeb,0x03,0x00,
		};
	memcpy ( curp, raw3, 96 );
	//ڸݼ
	qq_crypt( ENCRYPT, (unsigned char*)tmp, 416, 
		(unsigned char*)myqq.session_key, (unsigned char*)data_encrypted, &len ); //len = 432
	memcpy( data+16, data_encrypted, len );
	//׼
	cs_send_cmd( QQ_CMD_LOGIN, data, 16+len, 1 );
	DBG("2. LOGIN DATA PACKET SENT.%d", len+16);
	//wait for reply then.
	return 0;
}


//recv this after sending login info.
void login_recv_info( char* data, int len )
{
	//we got the packet now.
	//myqq.pass_encryted
	int len_ret;
	char tmp[512];
	len_ret = len;
	if( qq_crypt( DECRYPT, (unsigned char*)data, len, 
		(unsigned char*)myqq.pass_encrypted, (unsigned char*)tmp, &len_ret ) )
	{
		if( tmp[0] == 0 )
		{
			myqq.status = ST_ONLINE;
			login_save_info( tmp, len_ret );
			myqq_change_status( myqq.mode );
		}else{
			myqq.status = ST_ERROR;	//Unknown Error
		}
	}else{
		len_ret = len;
		if( qq_crypt( DECRYPT, (unsigned char*)data, len, 
			(unsigned char*)myqq.session_key, (unsigned char*)tmp, &len_ret ) )
		{
			if( tmp[0] == 0x01 ) //redirect.
			{
				myqq.status = ST_REDIRECT;
				login_redirect( tmp, len_ret );
			}else if ( tmp[0] == 0x02 || tmp[0] == 0x05 ){
				myqq.status = ST_WRPASS;	//Wrong Password
			}else{
				myqq.status = ST_ERROR;	//Unknown Error
			}
		}else{
			DBG("Can't decrypt data.");
			myqq.status = ST_ERROR;	//Unknown Error
		}
	}
}


//¼ӿ
EXPORT int myqq_login( uint uid, char* password, int hide )
{
	int t, ret;
	DBG("myqq_login");
	if( myqq.status == ST_OFFLINE )
	{
		myqq.seqno = rand() & 0xFFFF;
		myqq.sock_im = mysock_create( MYSOCK_UDP, NULL, 0 );
		if( myqq.server_ip && myqq.server_port )
		{
			DBG("USE REDIRECT IP AND PORT: %d", myqq.server_port);
			if( mysock_connect2( myqq.sock_im, myqq.server_ip, myqq.server_port ) < 0 )
				return -EFAULT;
		}else{
			if( mysock_connect( myqq.sock_im, "sz.tencent.com", 8000 ) < 0 )
				return -EFAULT;
		}
	}else{
		return 0;
	}
	myqq.status = ST_LOGGING;
	myqq.uid = uid;
	strncpy( myqq.password, password, MAX_PASSWORD );
	if( hide )
		myqq.mode = LM_HIDDEN;
	else
		myqq.mode = LM_NORMAL;
	MSG("Logging ...\n");
	if( login_get_token() < 0 )
	{
		MSG("Failed in getting login token.\n");
		myqq.status = ST_OFFLINE;
		return -EFAULT;
	}
	if( (ret=login_send_info())<0 )
	{
		DBG("Failed in getting login.");
		myqq.status = ST_OFFLINE;
		return -EFAULT;
	}
	cs_start_recv_packet();
	for( t=0; t<8; t++ )
	{
		switch( myqq.status ){
		case ST_ONLINE:
			goto LogComp;
		case ST_REDIRECT:
			DBG("Redirect Server.");
			mysock_close ( myqq.sock_im );
			myqq.status = ST_OFFLINE;
			cs_end_recv_packet();
			return myqq_login( myqq.uid, myqq.password, myqq.mode );
		case ST_WRPASS:
			DBG("Wrong Password.");
			MSG("Password is incorrect!\n");
			myqq.status = ST_OFFLINE;
			cs_end_recv_packet();
			return -EFAULT;
		case ST_LOGGING:
			break;
		default:
			myqq.status = ST_OFFLINE;
			cs_end_recv_packet();
			MSG("Login: Unknown Error.\n");
			return -EFAULT;
		}
//ֵһ!! Huang Guan 2008-02-02
#ifdef __WIN32__
		sleep(500);
#else
		sleep(1);
#endif
	}
LogComp:
	MSG("User Logged successfully.\n");
	buddy_get_info( myqq.uid );
	myqq_update_group_list();
	myqq_update_buddy_list();
	cs_start_keep_alive();
	return 0;
}

//Logout
EXPORT void myqq_logout()
{
	int i;
	// send four times as qq does.
	for (i = 0; i < 4; i++)
		cs_send_cmd_with_seqno( QQ_CMD_LOGOUT, myqq.pass_encrypted, 16, 0, 0xffff );
	//
	cs_end_keep_alive();
	cs_end_recv_packet();
	packet_list_cleanup();
	myqq.status = ST_OFFLINE;
}


//Change status
EXPORT void myqq_change_status( int mode )
{
	myqq.mode = mode;
	char data[5]; 	//Last four bytes are used for other status, like whether
				//there is a camera.
	memset( data, 0, 5 );	//zero all
	switch( mode&0xFF ){
	case LM_NORMAL:
		data[0] = QQ_BUDDY_ONLINE_NORMAL;
		break;
	case LM_HIDDEN:
		data[0] = QQ_BUDDY_ONLINE_INVISIBLE;
		break;
	case LM_AWAY:
		data[0] = QQ_BUDDY_ONLINE_AWAY;
		break;
	default:
		DBG("Unknown mode: %d", mode&0xff );
		data[0] = 0x0A;
	}
	//0A 1E 뿪40 
	cs_send_cmd( QQ_CMD_CHANGE_ONLINE_STATUS, data, 5, 1 );
}

//change status complete
void login_change_status_comp( char* data, int len )
{
	if( len!=1 )
	{
		DBG("error len!=1");
	}	
	else if( *data != 0x30 )
	{
		MSG("Failed to change status.");
	}
}


EXPORT int myqq_get_status()
{
	if( myqq.status == ST_OFFLINE )
		return LM_OFFLINE;
	return myqq.mode;	
}
