/*
 *  mysocket.c
 *
 *  Some functions to use the POSIX socket.
 *
 *  Copyright (C) 2008  Huang Guan
 *
 *  2008-01-31 Created.
 *
 *  Description: This file mainly includes the functions about 
 *               mysock_create, mysock_close
 *               mysock_send, mysock_recv
 *               mysock_connect, mysock_connect2
 */


#include <stdio.h>
#include <unistd.h>

#ifdef __WIN32__
#include <winsock.h>
#include <wininet.h>
#else
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>
#endif

#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include "debug.h"
#include "myqq.h"
#include "mysocket.h"

static void netaddr_set( char* name, struct sockaddr_in* addr );

//Create a socket TCP or UDP
int mysock_create( int type, char* ip, ushort port )
{
	int fd;
	struct sockaddr_in addr;
#ifdef __WIN32__
	WSADATA wsa_data;
	static int wsock_init = 0;
	if( !wsock_init )
	{
		int ret = WSAStartup(MAKEWORD(2,1), &wsa_data); //ʼWinSocket̬ӿ
		if( ret != 0 ) // ʼʧܣ
		{
			return -EFAULT;
		}
		DBG("WSA Startup.");
		wsock_init = 1;
	}
#endif
	switch( type )
	{
	case MYSOCK_TCP:
		fd = socket( PF_INET, SOCK_STREAM, 0 );
		memset( &addr, 0, sizeof(struct sockaddr_in) );
		addr.sin_family = PF_INET;
		if( ip )
			addr.sin_addr.s_addr = inet_addr( ip );
		else
			addr.sin_addr.s_addr = INADDR_ANY;
		addr.sin_port = htons( 0 );
		if( bind( fd, (struct sockaddr*)&addr, sizeof(struct sockaddr_in) ) < 0 )
		{
			DBG("bind socket error!");
			close( fd );
			return -EFAULT;
		}
		break;
	case MYSOCK_UDP:
		fd = socket( PF_INET, SOCK_DGRAM, 0 );
		memset( &addr, 0, sizeof(struct sockaddr) );
		addr.sin_family = PF_INET;
		if( ip )
			addr.sin_addr.s_addr = inet_addr( ip );
		else
			addr.sin_addr.s_addr = INADDR_ANY;
		addr.sin_port = htons( 0 );
		if( bind( fd, (struct sockaddr*)&addr, sizeof(struct sockaddr_in) ) < 0 )
		{
			DBG("bind socket error!");
			close( fd );
			return -EFAULT;
		}
		break;
	}
	return fd;
}

//close a socket
void mysock_close( int fd )
{
	close( fd );
} 

//connect a socket to remote server
int mysock_connect( int fd, char* ip, ushort port )
{
	struct sockaddr_in addr;
	memset( &addr, 0, sizeof(struct sockaddr_in) );
	addr.sin_family = PF_INET;
	netaddr_set( ip, &addr );
	addr.sin_port = htons( port );
	if( connect( fd, (struct sockaddr*)&addr, sizeof(struct sockaddr_in)) < 0 )
	{
		DBG("mysock connect failed.");
		return -EFAULT;
	}
	return 0;
}

//connect a socket to remote server
int mysock_connect2( int fd, uint ip, ushort port )
{
	struct sockaddr_in addr;
	memset( &addr, 0, sizeof(struct sockaddr_in) );
	addr.sin_family = PF_INET;
	addr.sin_addr.s_addr = htonl( ip );
	addr.sin_port = htons( port );
	if( connect( fd, (struct sockaddr*)&addr, sizeof(struct sockaddr_in)) < 0 )
	{
		DBG("mysock connect failed.");
		return -EFAULT;
	}
	return 0;
}

static void netaddr_set( char* name, struct sockaddr_in* addr )
{
	if( (addr->sin_addr.s_addr = inet_addr( name ) ) == -1 )
	{
		//it's not an IP.
		//not an ip, maybe a domain
		struct hostent *host;
		host = gethostbyname( name );
		if( host )
		{
			addr->sin_addr.s_addr = *(size_t*) host->h_addr_list[0];
			DBG("Get IP: %s", inet_ntoa( addr->sin_addr ) );
			return;
		}
		DBG("Failed to get ip by %s", name );
	}	
}

int mysock_send( int fd, char* buf, size_t size )
{
	int ret;
#ifdef __WIN32__
	ret = send( fd, buf, size, 0 );
#else
	ret = write( fd, buf, size );
#endif
	return ret;
}

int mysock_recv( int fd, char* buf, size_t size )
{
	int ret;
#ifdef __WIN32__
	ret = recv( fd, buf, size, 0 );
#else
	ret = read( fd, buf, size );
#endif
	return ret;
}


